# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""JWT utilities."""

from datetime import datetime, timezone

import jwt


def is_token_expired(token: str) -> bool:
    """Return True if the given token is expired."""
    try:
        decoded_token = jwt.decode(token, options={"verify_signature": False})
    except jwt.DecodeError:
        return True

    expiration_date = decoded_token.get("exp", 0)

    # NOTE: ``datetime.utcnow`` doesn't have timezone info, so we use ``datetime.now``
    return expiration_date < datetime.now(tz=timezone.utc).timestamp()
